"""
File Catalog
============

Interface to communicate with a File Catalog using REST over HTTP.
"""

import logging

import requests

from .jsonUtil import json_encode, json_decode

logger = logging.getLogger('file_catalog')


class FileCatalog(object):
    """
    High level file catalog interface.

    Args:
        url (str): url of the file catalog server
        overwrite (bool): overwrite entries when adding new files
    """
    def __init__(self, url, overwrite=False):
        self.fc = FileCatalogLowLevel(url)
        self.overwrite = overwrite

    def add(self, name, path, checksum, metadata=None, overwrite=None):
        """
        Add a file to the catalog.

        Args:
            name (str): unique name of file
            path (str): url of file
            checksum (str): sha512 checksum of file
            metadata (dict): (optional) additional metadata
            overwrite (bool): overwrite entries when adding new files
        """
        if not metadata:
            metadata = {}
        if overwrite is None:
            overwrite = self.overwrite
        metadata.update({'uid': name, 'checksum': checksum,
                         'locations': [path]})
        try:
            self.fc[name] = metadata
        except requests.exceptions.HTTPError:
            if overwrite:
                del self.fc[name]
                self.fc[name] = metadata
            else:
                raise

    def get(self, name):
        """
        Get a file path and checksum from the catalog.

        Args:
            name (str): unique name of file

        Returns:
            tuple: path, checksum
        """
        m = self.fc[name]
        return (m['locations'][0], m['checksum'])

    def get_metadata(self, name):
        """
        Get all file metadata from the catalog.

        Args:
            name (str): unique name of file

        Returns:
            dict: metadata information
        """
        return self.fc[name]

    def delete(self, name):
        """
        Remove a file from the catalog

        Args:
            name (str): unique name of file
        """
        del self.fc[name]


class FileCatalogLowLevel(object):
    """
    Low level file catalog interface.  Use like a dict::

        fc = FileCatalog('http://file_catalog.com')
        fc['my_new_file'] = {'locations':['/this/is/a/path']}

    Args:
        url (str): url of the file catalog server
        timeout (float): (optional) seconds to wait for a query to finish
    """
    def __init__(self, url, timeout=60):
        self.url = url
        self.timeout = timeout
        self.session = requests.Session()

    def _getfileurl(self, uid):
        for _ in range(5):
            try:
                r = self.session.get(self.url+'/api/files',
                                     params={'query':json_encode({'uid':uid})},
                                     timeout=self.timeout)
            except requests.exceptions.Timeout:
                continue
            if r.status_code == 429:
                continue
            r.raise_for_status()
            files = json_decode(r.text)['files']
            break
        else:
            raise Exception('server is too busy')
        if len(files) != 1:
            raise KeyError()
        return self.url+files[0]

    def __getitem__(self, uid):
        url = self._getfileurl(uid)
        for _ in range(5):
            try:
                r = self.session.get(url, timeout=self.timeout)
            except requests.exceptions.Timeout:
                continue
            if r.status_code == 429:
                continue
            r.raise_for_status()
            return json_decode(r.text)
        raise Exception('server is too busy')

    def __setitem__(self, uid, value):
        meta = value.copy()
        meta['uid'] = uid
        data = json_encode(meta)
        try:
            url = self._getfileurl(uid)
        except KeyError:
            # does not exist
            method = self.session.post
            url = self.url+'/api/files'
        else:
            # exists, so update
            method = self.session.put
        for _ in range(5):
            try:
                r = method(url, data=data,
                           timeout=self.timeout)
            except requests.exceptions.Timeout:
                continue
            if r.status_code == 429:
                continue
            r.raise_for_status()
            return
        raise Exception('server is too busy')

    def __delitem__(self, uid):
        url = self._getfileurl(uid)
        for _ in range(5):
            try:
                r = self.session.delete(url, timeout=self.timeout)
            except requests.exceptions.Timeout:
                continue
            if r.status_code == 429:
                continue
            r.raise_for_status()
            return
        raise Exception('server is too busy')



import logging
import json
import uuid
from collections import defaultdict

import pymongo
import tornado.web

from ..base_handler import APIBase
from ..auth import authorization, attr_auth
from iceprod.server.util import nowstr
from iceprod.server.states import JOB_STATUS, JOB_STATUS_START, job_prev_statuses, job_status_sort

logger = logging.getLogger('rest.jobs')


def setup(handler_cfg):
    """
    Setup method for Jobs REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            (r'/jobs', MultiJobsHandler, handler_cfg),
            (r'/jobs/(?P<job_id>\w+)', JobsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/jobs', DatasetMultiJobsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/jobs/(?P<job_id>\w+)', DatasetJobsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/jobs/(?P<job_id>\w+)/status', DatasetJobsStatusHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/job_actions/bulk_status/(?P<status>\w+)', DatasetJobBulkStatusHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/job_actions/bulk_suspend', DatasetJobBulkSuspendHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/job_actions/bulk_reset', DatasetJobBulkResetHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/job_actions/bulk_hard_reset', DatasetJobBulkHardResetHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/job_summaries/status', DatasetJobSummariesStatusHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/job_counts/status', DatasetJobCountsStatusHandler, handler_cfg),
        ],
        'database': 'jobs',
        'indexes': {
            'jobs': {
                'job_id_index': {'keys': 'job_id', 'unique': True},
                'dataset_id_index': {'keys': 'dataset_id', 'unique': False},
            }
        }
    }


class MultiJobsHandler(APIBase):
    """
    Handle multi jobs requests.
    """
    @authorization(roles=['admin', 'system'])
    async def post(self):
        """
        Create a job entry.

        Body should contain the job data.

        Returns:
            dict: {'result': <job_id>}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'dataset_id': str,
            'job_index': int,
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        opt_fields = {
            'job_id': str,
            'status': str,
        }
        for k in opt_fields:
            if k in data and not isinstance(data[k], opt_fields[k]):
                r = 'key "{}" should be of type {}'.format(k, opt_fields[k].__name__)
                raise tornado.web.HTTPError(400, reason=r)

        bad_fields = set(data).difference(set(opt_fields).union(req_fields))
        if bad_fields:
            r = 'invalid keys found'
            raise tornado.web.HTTPError(400, reason=r)

        if 'job_id' in data:
            try:
                uuid.UUID(hex=data['job_id']).hex
            except Exception:
                raise tornado.web.HTTPError(400, reason='job_id should be a valid uuid')

        if 'status' in data and data['status'] not in JOB_STATUS:
            raise tornado.web.HTTPError(400, reason='unknown status')

        # set some fields
        if 'job_id' not in data:
            data['job_id'] = uuid.uuid1().hex
        if 'status' not in data:
            data['status'] = JOB_STATUS_START
        data['status_changed'] = nowstr()

        await self.db.jobs.insert_one(data)
        self.set_status(201)
        self.write({'result': data['job_id']})
        self.finish()


class JobsHandler(APIBase):
    """
    Handle single job requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, job_id):
        """
        Get a job entry.

        Args:
            job_id (str): the job id

        Returns:
            dict: job entry
        """
        ret = await self.db.jobs.find_one({'job_id':job_id}, projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Job not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin', 'system'])
    async def patch(self, job_id):
        """
        Update a job entry.

        Body should contain the job data to update.  Note that this will
        perform a merge (not replace).

        Args:
            job_id (str): the job id

        Returns:
            dict: updated job entry
        """
        data = json.loads(self.request.body)
        if not data:
            raise tornado.web.HTTPError(400, reason='Missing update data')

        ret = await self.db.jobs.find_one_and_update(
            {'job_id':job_id},
            {'$set':data},
            projection={'_id':False},
            return_document=pymongo.ReturnDocument.AFTER
        )
        if not ret:
            self.send_error(404, reason="Job not found")
        else:
            self.write(ret)
            self.finish()


class DatasetMultiJobsHandler(APIBase):
    """
    Handle multi jobs requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id):
        """
        Get all jobs for a dataset.

        Params (optional):
            job_index: job_index to filter by
            status: | separated list of job status to filter by
            keys: | separated list of keys to return for each task

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {'job_id':{job_data}}
        """
        filters = {'dataset_id':dataset_id}
        status = self.get_argument('status', None)
        if status:
            status_list = status.split('|')
            if any(s not in JOB_STATUS for s in status_list):
                raise tornado.web.HTTPError(400, reason='Unknown status')
            filters['status'] = {'$in': status_list}

        job_index = self.get_argument('job_index', None)
        if job_index:
            try:
                filters['job_index'] = int(job_index)
            except ValueError:
                raise tornado.web.HTTPError(400, reason='Bad argument "job_index": must be integer')

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
            projection['job_id'] = True

        cursor = self.db.jobs.find(filters, projection=projection)
        ret = {}
        async for row in cursor:
            ret[row['job_id']] = row
        self.write(ret)
        self.finish()


class DatasetJobsHandler(APIBase):
    """
    Handle single job requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id, job_id):
        """
        Get a job entry.

        Args:
            dataset_id (str): dataset id
            job_id (str): the job id

        Returns:
            dict: job entry
        """
        ret = await self.db.jobs.find_one(
            {'job_id':job_id,'dataset_id':dataset_id},
            projection={'_id':False}
        )
        if not ret:
            self.send_error(404, reason="Job not found")
        else:
            self.write(ret)
            self.finish()


class DatasetJobsStatusHandler(APIBase):
    """
    Handle single job requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def put(self, dataset_id, job_id):
        """
        Set a job status, following possible state transitions.

        Body should have {'status': <new_status>}

        Args:
            dataset_id (str): dataset id
            job_id (str): the job id

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'status' not in data:
            raise tornado.web.HTTPError(400, reason='Missing status in body')
        if data['status'] not in JOB_STATUS:
            raise tornado.web.HTTPError(400, reason='Bad status')
        update_data = {
            'status': data['status'],
            'status_changed': nowstr(),
        }

        ret = await self.db.jobs.update_one(
            {'job_id': job_id, 'dataset_id': dataset_id, 'status': {'$in': job_prev_statuses(data['status'])}},
            {'$set': update_data}
        )
        if (not ret) or ret.modified_count < 1:
            ret = await self.db.jobs.find_one({'job_id': job_id, 'dataset_id': dataset_id})
            if not ret:
                self.send_error(404, reason="Job not found")
                return
            elif ret['status'] != data['status']:
                self.send_error(400, reason="Bad state transition for status")
                return

        self.write({})
        self.finish()


class DatasetJobBulkStatusHandler(APIBase):
    """
    Update the status of multiple jobs at once.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def post(self, dataset_id, status):
        """
        Set multiple jobs' status, following possible state transitions.

        Body should have {'jobs': [<job_id>, <job_id>, ...]}

        Args:
            dataset_id (str): dataset id
            status (str): the status

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'jobs' not in data or not data['jobs']:
            raise tornado.web.HTTPError(400, reason='Missing jobs in body')
        jobs = list(data['jobs'])
        if len(jobs) > 100000:
            raise tornado.web.HTTPError(400, reason='Too many jobs specified (limit: 100k)')
        if status not in JOB_STATUS:
            raise tornado.web.HTTPError(400, reason='Bad status')
        query = {
            'dataset_id': dataset_id,
            'job_id': {'$in': jobs},
            'status': {'$in': job_prev_statuses(status)}
        }
        update_data = {
            'status': status,
            'status_changed': nowstr(),
        }

        ret = await self.db.jobs.update_many(query, {'$set': update_data})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Jobs not found")
        else:
            self.write({})
            self.finish()


class DatasetJobBulkSuspendHandler(APIBase):
    """
    Suspend jobs in a dataset.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def post(self, dataset_id):
        """
        Do a suspend on jobs.

        Body may have {'jobs': [<job_id>, <job_id>, ...]}
        If it does not, all jobs in a dataset are suspended.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: empty dict
        """
        query = {
            'dataset_id': dataset_id,
            'status': {'$in': job_prev_statuses('suspended')},
        }
        update_data = {
            'status': 'suspended',
            'status_changed': nowstr(),
        }

        if self.request.body:
            data = json.loads(self.request.body)
            if data and 'jobs' in data:
                jobs = list(data['jobs'])
                if len(jobs) > 100000:
                    raise tornado.web.HTTPError(400, reason='Too many jobs specified (limit: 100k)')
                query['job_id'] = {'$in': jobs}

        await self.db.jobs.update_many(query, {'$set': update_data})
        self.write({})
        self.finish()


class DatasetJobBulkResetHandler(APIBase):
    """
    Reset jobs in a dataset.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def post(self, dataset_id):
        """
        Do a reset on jobs.

        Body may have {'jobs': [<job_id>, <job_id>, ...]}
        If it does not, all jobs in a dataset are reset.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: empty dict
        """
        query = {
            'dataset_id': dataset_id,
            'status': {'$in': job_prev_statuses(JOB_STATUS_START)},
        }
        update_data = {
            'status': JOB_STATUS_START,
            'status_changed': nowstr(),
        }

        if self.request.body:
            data = json.loads(self.request.body)
            if data and 'jobs' in data:
                jobs = list(data['jobs'])
                if len(jobs) > 100000:
                    raise tornado.web.HTTPError(400, reason='Too many jobs specified (limit: 100k)')
                query['job_id'] = {'$in': jobs}

        logger.info('bulk reset query: %r', query)
        await self.db.jobs.update_many(query, {'$set': update_data})
        self.write({})
        self.finish()


class DatasetJobBulkHardResetHandler(APIBase):
    """
    Hard reset jobs in a dataset.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def post(self, dataset_id):
        """
        Do a hard reset on jobs.

        Body may have {'jobs': [<job_id>, <job_id>, ...]}
        If it does not, all jobs in a dataset are hard reset.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: empty dict
        """
        query = {
            'dataset_id': dataset_id,
        }
        update_data = {
            'status': JOB_STATUS_START,
            'status_changed': nowstr(),
        }

        if self.request.body:
            data = json.loads(self.request.body)
            if data and 'jobs' in data:
                jobs = list(data['jobs'])
                if len(jobs) > 100000:
                    raise tornado.web.HTTPError(400, reason='Too many jobs specified (limit: 100k)')
                query['job_id'] = {'$in': jobs}

        await self.db.jobs.update_many(query, {'$set': update_data})
        self.write({})
        self.finish()


class DatasetJobSummariesStatusHandler(APIBase):
    """
    Handle job summary grouping by status.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id):
        """
        Get the job summary for all jobs in a dataset, group by status.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<status>: [<job_id>,]}
        """
        cursor = self.db.jobs.find(
            {'dataset_id':dataset_id},
            projection={'_id':False,'status':True,'job_id':True}
        )
        ret = defaultdict(list)
        async for row in cursor:
            ret[row['status']].append(row['job_id'])
        ret2 = {}
        for k in sorted(ret, key=job_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()


class DatasetJobCountsStatusHandler(APIBase):
    """
    Handle job counts by status.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id):
        """
        Get the job counts for all jobs in a dataset, group by status.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<status>: [<job_id>,]}
        """
        cursor = self.db.jobs.aggregate([
            {'$match': {'dataset_id': dataset_id}},
            {'$group': {'_id':'$status', 'total': {'$sum':1}}},
        ])
        ret = {}
        async for row in cursor:
            ret[row['_id']] = row['total']
        ret2 = {}
        for k in sorted(ret, key=job_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()



"""Utility functions"""
from datetime import datetime, UTC


def datetime2str(dt):
    """Convert a datetime object to ISO 8601 string"""
    return dt.strftime('%Y-%m-%dT%H:%M:%S.%f')


def nowstr():
    """Get an ISO 8601 string of the current time in UTC"""
    return datetime.now(UTC).strftime('%Y-%m-%dT%H:%M:%S.%f')


def str2datetime(st):
    """Convert a ISO 8601 string to datetime object"""
    if '.' in st:
        return datetime.strptime(st, "%Y-%m-%dT%H:%M:%S.%f")
    else:
        return datetime.strptime(st, "%Y-%m-%dT%H:%M:%S")



"""
Clean job temp directories.

Check job temp directories, and if the job is complete then delete it.
"""

import argparse
import asyncio
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from functools import partial
import logging
import os
from urllib.parse import urlparse
from wipac_dev_tools import from_environment

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.core.gridftp import GridFTP
from iceprod.s3 import S3
from iceprod.server.util import str2datetime

logger = logging.getLogger('job_temp_cleaning')


async def list_dataset_job_dirs_gridftp(path, prefix=None, executor=None):
    dataset_dirs = defaultdict(dict)

    async def list_job_dirs(d):
        jobs = await asyncio.wrap_future(executor.submit(partial(GridFTP.list, os.path.join(path, d), details=True)))
        for entry2 in jobs:
            if not entry2.directory:
                continue
            j = entry2.name
            dataset_dirs[d][j] = entry2.size

    if prefix:
        await list_job_dirs(prefix)
    else:
        dirs = await asyncio.wrap_future(executor.submit(partial(GridFTP.list, path, details=True)))
        for entry in dirs:
            if not entry.directory:
                continue
            try:
                await list_job_dirs(entry.name)
            except Exception:
                logger.warning('error listing %s', entry.name, exc_info=True)
    return dataset_dirs


async def rmtree_gridftp(path, executor=None):
    await asyncio.wrap_future(executor.submit(partial(GridFTP.rmtree, path)))


async def list_dataset_job_dirs_webdav(path, prefix=None, rest_client=None):
    dataset_dirs = defaultdict(dict)
    return dataset_dirs


async def rmtree_webdav(path, rest_client=None):
    pass


async def list_dataset_job_dirs_s3(path, prefix=None, s3_client=None):
    if prefix:
        path = os.path.join(path, prefix)
        ret = await s3_client.list(path)
        return {prefix: ret}
    else:
        return await s3_client.list(path)


async def rmtree_s3(path, s3_client=None):
    await s3_client.rmtree(path)


async def run(rest_client, temp_dir, list_dirs, rmtree, dataset=None, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        temp_dir (str): temporary directory to clean
        list_dirs (callable): list temporary directory
        rmtree (callable): delete temporary directory and contents
        dataset (int): dataset num to run over (optional)
        debug (bool): debug flag to propagate exceptions
    """
    suspend_time = timedelta(days=90)
    now = datetime.utcnow()

    try:
        # get all the job_indexes currently in tmp
        if dataset:
            dataset_dirs = await list_dirs(temp_dir, prefix=str(dataset))
        else:
            dataset_dirs = await list_dirs(temp_dir)
        logger.debug('dataset_dirs: %r', dataset_dirs)

        ret = await rest_client.request('GET', '/datasets?keys=dataset_id|dataset')
        logger.debug('datasets api raw: %r', ret)
        datasets = {}
        for d in ret:
            datasets[str(ret[d]['dataset'])] = ret[d]['dataset_id']
        logger.debug('datasets: %r', datasets)

        for d in dataset_dirs:
            if not isinstance(dataset_dirs[d], dict):
                continue
            logger.info('temp cleaning for dataset %r', d)
            job_dirs = dataset_dirs[d]
            logger.debug('job_dirs: %r', job_dirs)

            try:
                jobs = await rest_client.request('GET', f'/datasets/{datasets[d]}/jobs', {'keys':'status|status_changed|job_index'})
            except Exception:
                logger.error('failed to get jobs for dataset %r', d, exc_info=True)
                if debug:
                    raise
                continue
            logger.debug('jobs: %r', jobs)

            job_indexes = set()
            for job in jobs.values():
                if job['status'] == 'complete' or (job['status'] != 'processing' and now - str2datetime(job['status_changed']) > suspend_time):
                    job_indexes.add(job['job_index'])
            logger.debug('job_indexes: %r', job_indexes)

            futures = set()
            for j in job_dirs:
                if isinstance(job_dirs[j], dict):
                    continue
                if not j.isnumeric():
                    if debug:
                        logger.info('j is not numeric: %r', j)
                        raise Exception('not numeric')
                    continue
                if int(j) in job_indexes:
                    while len(futures) >= 16:
                        done, futures = await asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
                        for f in done:
                            try:
                                await f
                            except Exception:
                                logger.warning('failed to clean site_temp', exc_info=True)
                                if debug:
                                    raise

                    dagtemp = os.path.join(temp_dir, d, j)
                    logger.info('cleaning site_temp %r', dagtemp)
                    futures.add(asyncio.create_task(rmtree(dagtemp)))

            if futures:
                for f in asyncio.as_completed(futures):
                    try:
                        await f
                    except Exception:
                        logger.warning('failed to clean site_temp', exc_info=True)
                        if debug:
                            raise

    except Exception:
        logger.error('error checking job temp', exc_info=True)
        if debug:
            raise


def main():
    default_config = {
        'SITE_TEMP': 'gsiftp://gridftp-scratch.icecube.wisc.edu/mnt/tank/simprod',
        'S3_ACCESS_KEY': '',
        'S3_SECRET_KEY': '',
    }
    config = from_environment(default_config)

    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('-d', '--dataset', type=str, help='dataset num (optional)')
    parser.add_argument('--site-temp', default=config['SITE_TEMP'], help='site temp location')
    parser.add_argument('--s3-access-key', default=config['S3_ACCESS_KEY'], help='s3 access key')
    parser.add_argument('--s3-secret-key', default=config['S3_SECRET_KEY'], help='s3 secret key')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    if args.site_temp.startswith('gsiftp://'):
        logging.info('using GridFTP')
        pool = ThreadPoolExecutor()
        listdir = partial(list_dataset_job_dirs_gridftp, executor=pool)
        rmtree = partial(rmtree_gridftp, executor=pool)
    elif args.s3_access_key:
        logging.info('using S3')
        o = urlparse(args.site_temp)
        url = o.path.lstrip('/')
        if '/' in url:
            bucket, url = url.split('/', 1)
        else:
            bucket = url
            url = ''
        host = f'{o.scheme}://{o.netloc}'
        logging.info('host: %r, bucket: %r', host, bucket)
        args.site_temp = url
        s3client = S3(host, bucket=bucket, access_key=args.s3_access_key, secret_key=args.s3_secret_key)
        listdir = partial(list_dataset_job_dirs_s3, s3_client=s3client)
        rmtree = partial(rmtree_s3, s3_client=s3client)
    else:
        raise RuntimeError('unknown type of scratch')

    logging.info('temp dir: %r', args.site_temp)
    asyncio.run(run(rest_client, args.site_temp, list_dirs=listdir, rmtree=rmtree, dataset=args.dataset, debug=args.debug))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Add a site to the master for a pool
"""

from __future__ import absolute_import, division, print_function

import os
import sys
from functools import partial

import iceprod.server.basic_config
import iceprod.server.RPCinternal


def main(cfgfile):
    cfg = iceprod.server.basic_config.BasicConfig()
    cfg.read_file(cfgfile)
    class Response:
        pass
    kwargs = {
        'address':cfg.messaging_url,
        'block':False,
        'service_name':'add_site',
        'service_class':Response(),
        'async':False,
    }
    messaging = iceprod.server.RPCinternal.RPCService(**kwargs)
    messaging.start()

    site_id = raw_input('site_id:')

    passkey = messaging.db.add_site_to_master(site_id=site_id)
    if isinstance(passkey,Exception):
        raise passkey
    print('passkey:',passkey)

if __name__ == '__main__':
    import argparse

    # Override values with cmdline options
    parser = argparse.ArgumentParser(description='IceProd Server')
    parser.add_argument('-f','--file',dest='file',type=str,default=None,
                        help='Path to config file')
    args = parser.parse_args()

    if args.file:
        cfgfile = os.path.expanduser(os.path.expandvars(args.file))
    else:
        cfgfile = iceprod.server.basic_config.locateconfig()

    main(cfgfile)



"""
Functions used to calculate dataset and task priority.
"""
import argparse
import asyncio
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.roles_groups import GROUP_PRIORITIES

logger = logging.getLogger('priority')


class Priority:
    def __init__(self, rest_client):
        self.rest_client = rest_client
        self.dataset_cache = {}
        self.user_cache = {}
        self.group_cache = {}

    async def _populate_dataset_cache(self):
        args = {
            'keys': 'dataset_id|priority|jobs_submitted|tasks_submitted|group|username',
            'status': 'processing',
        }
        self.dataset_cache = await self.rest_client.request('GET', '/datasets', args)
        dataset_ids = list(self.dataset_cache)
        args = {
            'keys': 'task_id|dataset_id|job_index|task_index',
            'status': 'idle|waiting|queued|processing',
        }
        futures = set()
        while dataset_ids:
            if len(futures) >= 20:
                done, pending = await asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
                futures = pending
                for f in done:
                    ret = await f
                    if ret:
                        d = list(ret.values())[0]['dataset_id']
                        self.dataset_cache[d]['tasks'] = {
                            k: {'task_index': ret[k]['task_index'], 'job_index': ret[k]['job_index']}
                            for k in ret
                        }

            dataset_id = dataset_ids.pop()
            self.dataset_cache[dataset_id]['tasks'] = {}
            t = asyncio.create_task(self.rest_client.request('GET', f'/datasets/{dataset_id}/tasks', args))
            futures.add(t)
        while futures:
            done, pending = await asyncio.wait(futures)
            futures = pending
            for f in done:
                ret = await f
                try:
                    d = ret.pop('dataset_id')
                except KeyError:
                    continue
                self.dataset_cache[d]['tasks'] = ret

    async def _populate_dataset_task_cache(self, dataset_id, task_id):
        if dataset_id not in self.dataset_cache:
            return
        if task_id not in self.dataset_cache[dataset_id]['tasks']:
            args = {
                'keys': 'task_id|job_index|task_index',
            }
            ret = await self.rest_client.request('GET', f'/datasets/{dataset_id}/tasks/{task_id}', args)
            self.dataset_cache[dataset_id]['tasks'][task_id] = ret

    async def _populate_user_cache(self):
        ret = await self.rest_client.request('GET', '/users')
        for user in ret['results']:
            if 'priority' not in user:
                user['priority'] = 0.5
            self.user_cache[user['username']] = user

    async def _get_dataset(self, dataset_id):
        if not self.dataset_cache:
            await self._populate_dataset_cache()
        return self.dataset_cache[dataset_id]

    async def _get_max_dataset_prio_user(self, user):
        if not self.dataset_cache:
            await self._populate_dataset_cache()
        try:
            return max(d['priority'] for d in self.dataset_cache.values() if 'priority' in d and d['username'] == user)
        except ValueError:
            return 1.

    async def _get_max_dataset_prio_group(self, group):
        if not self.dataset_cache:
            await self._populate_dataset_cache()
        try:
            return max(d['priority'] for d in self.dataset_cache.values() if 'priority' in d and d['group'] == group)
        except ValueError:
            return 1.

    async def _get_user_prio(self, user):
        if not self.user_cache:
            await self._populate_user_cache()
        if user not in self.user_cache or 'priority' not in self.user_cache[user]:
            return 0.5
        return self.user_cache[user]['priority']

    async def _get_group_prio(self, group):
        return GROUP_PRIORITIES.get(group, 0.5)

    async def _get_num_tasks(self, dataset_id=None):
        if not self.dataset_cache:
            await self._populate_dataset_cache()
        num = 0
        for d in self.dataset_cache:
            if dataset_id is None or dataset_id == d:
                try:
                    num += len(self.dataset_cache[d]['tasks'])
                except ValueError:
                    pass
        return num

    async def get_dataset_prio(self, dataset_id):
        """
        Calculate priority for a dataset.

        Args:
            dataset_id: dataset id

        Returns:
            float: priority between 0 and 1
        """
        try:
            dataset = await self._get_dataset(dataset_id)
        except KeyError:
            logger.warning(f'cannot find dataset {dataset_id}', exc_info=True)
            return 0.

        dataset_prio = dataset['priority']
        user = dataset['username']
        group = dataset['group']
        logger.debug(f'{dataset_id} dataset_prio: {dataset_prio}')
        max_dataset_prio = await self._get_max_dataset_prio_user(user)
        logger.debug(f'{dataset_id} max_dataset_prio: {max_dataset_prio}')
        max_dataset_prio_group = await self._get_max_dataset_prio_group(group)
        logger.debug(f'{dataset_id} max_dataset_prio_group: {max_dataset_prio_group}')

        if group == 'users':
            user_prio = await self._get_user_prio(user)
            logger.debug(f'{dataset_id} user_prio: {user_prio}')
            group_prio = await self._get_group_prio(group)
            logger.debug(f'{dataset_id} group_prio: {group_prio}')
        else:
            user_prio = 1.0
            group_prio = await self._get_group_prio(group)
            logger.debug(f'{dataset_id} group_prio: {group_prio}')

        num_dataset_jobs = dataset['jobs_submitted']
        logger.debug(f'{dataset_id} num_dataset_jobs: {num_dataset_jobs}')

        num_all_tasks = await self._get_num_tasks()
        logger.debug(f'{dataset_id} num_all_tasks: {num_all_tasks}')
        num_dataset_tasks = dataset['tasks_submitted']
        logger.debug(f'{dataset_id} num_dataset_tasks: {num_dataset_tasks}')
        num_dataset_tasks_avail = await self._get_num_tasks(dataset_id)
        logger.debug(f'{dataset_id} num_dataset_tasks_avail: {num_dataset_tasks_avail}')

        # general priority
        priority = 1.
        if max_dataset_prio > 0:
            priority *= dataset_prio / max_dataset_prio
            logger.info(f'{dataset_id} after dataset adjustment: {priority}')
        if group == 'users':
            logger.info('group is "users", so skipping dataset group adjustment')
        elif max_dataset_prio_group > 0:
            priority -= (1. - dataset_prio / max_dataset_prio_group) / 4.
            logger.info(f'{dataset_id} after dataset group adjustment: {priority}')
        priority *= user_prio**.5
        logger.info(f'{dataset_id} after user adjustment: {priority}')
        priority *= group_prio
        logger.info(f'{dataset_id} after group adjustment: {priority}')

        # bias against large datasets
        factor = (10000. / num_dataset_jobs)**.05 if num_dataset_jobs > 0 else 1.
        if factor > 1:
            factor = 1.
        priority *= factor
        logger.info(f'{dataset_id} after large dataset adjustment: {priority}')

        priority -= (1. * num_dataset_tasks_avail / num_all_tasks) / 4.
        logger.info(f'{dataset_id} after avail tasks adjustment: {priority}')

        if priority < 0.:
            priority = 0.
        elif priority > 1.:
            priority = 1.
        logger.info(f'{dataset_id} final priority: {priority}')

        return priority

    async def get_task_prio(self, dataset_id, task_id):
        """
        Calculate priority for a task.

        Args:
            dataset_id: dataset id
            task_id: task id

        Returns:
            float: priority between 0 and 1
        """
        try:
            dataset = await self._get_dataset(dataset_id)
        except Exception:
            logger.warning(f'cannot find dataset {dataset_id}', exc_info=True)
            return 0.
        if dataset['tasks_submitted'] < 1:
            return 0.

        await self._populate_dataset_task_cache(dataset_id, task_id)

        priority = await self.get_dataset_prio(dataset_id)

        task = dataset['tasks'][task_id]
        tasks_per_job = dataset['tasks_submitted'] / dataset['jobs_submitted']

        # bias towards finishing jobs
        priority += (1. * task['task_index'] / tasks_per_job) / 10.
        logger.info(f'{dataset_id}.{task_id} after finishing jobs adjustment: {priority}')

        # bias towards first jobs in dataset
        priority += (1. * (dataset['jobs_submitted'] - task['job_index']) / dataset['jobs_submitted']) / 100.
        logger.info(f'{dataset_id}.{task_id} after first jobs adjustment: {priority}')

        # boost towards first 100 jobs (or small datasets)
        if task['job_index'] < 100:
            priority += (100. - task['job_index']) / 100.
            logger.info(f'{dataset_id}.{task_id} after initial 100 jobs adjustment: {priority}')

        if priority < 0.:
            priority = 0.
        elif priority > 1.:
            priority = 1.
        logger.info(f'{dataset_id}.{task_id} final priority: {priority}')

        return priority


def main():
    parser = argparse.ArgumentParser(description='get priority')
    add_auth_to_argparse(parser)
    parser.add_argument('--debug', default=False, type=bool, help='debug enabled/disabled')
    parser.add_argument('dataset_id', help='dataset_id')
    parser.add_argument('--task_id', help='task_id (optional)')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=logging.DEBUG if args.debug else logging.INFO)

    rest_client = create_rest_client(args)

    p = Priority(rest_client)
    if args.task_id:
        ret = asyncio.run(p.get_task_prio(args.dataset_id, args.task_id))
        print('task priority', ret)
    else:
        ret = asyncio.run(p.get_dataset_prio(args.dataset_id))
        print('dataset priority', ret)


if __name__ == '__main__':
    main()



"""
Manage resources like CPUs and Memory.  Default values, detection, tracking,
policy, etc.
"""

import os
import time
import math
from copy import deepcopy

import subprocess
from collections import deque, defaultdict, OrderedDict
import logging

try:
    import psutil
except ImportError:
    psutil = None

try:
    import classad
except ImportError:
    def classad_to_dict(text):
        ret = {}
        for line in text.split('\n'):
            if (not line) or '=' not in line:
                continue
            parts = line.split('=',1)
            ret[parts[0].strip().lower()] = parts[1].strip()
        return ret
else:
    def classad_to_dict(text):
        ret = {}
        c = classad.parseOne(text)
        for k in c.keys():
            try:
                ret[k.lower()] = c.eval(k)
            except TypeError:
                ret[k.lower()] = c[k]
        return ret

# default rounding bins for resources
RESOURCE_BINS = {
    'cpu': list(range(1, 1000)),
    'gpu': list(range(0, 100)),
    'memory': [x/10. for x in list(range(5, 50, 5)) + list(range(50, 200, 10)) + list(range(200, 1000, 40)) + list(range(1000, 10000, 100))],
    'disk': list(range(1, 10)) + list(range(10, 50, 2)) + list(range(50, 100, 5)) + list(range(100, 2000, 20)),
    'time': [x/60. for x in list(range(10, 60, 10)) + list(range(60, 360, 15)) + list(range(360, 1440, 60)) + list(range(1440, 20240, 240))],
}


class Resources:
    """
    Task (and node) resource definition and tracking.
    """

    # ordering based on priority for matching
    defaults = OrderedDict([
        ('gpu', []),
        ('memory', 1.0),
        ('disk', 10.0),
        ('time', 1.0),
        ('cpu', 1),
    ])
    """Default resource values

    cpu
        integer

    gpu
        fungible - a list of unique items

    memory
        float, in GB

    disk
        float, in GB

    time
        float, in hours
    """

    overusage_limits = {
        'cpu': {'ignore': 2., 'allowed': 4., 'movingavg':10},
        'gpu': {'ignore': 1.5, 'allowed': 1.5, 'movingavg':10},
        'memory': {'ignore': 0, 'allowed': 10.},
        'disk': {'ignore': 0, 'allowed': 10.},
        'time': {'ignore': 0, 'allowed': 10.},
    }
    """Overusage limits

    ignore
        Ignore extra usage up to this ratio over the allocated resource.

    allowed
        Allow extra usage if there are available resources, up to this ratio.

    movingavg
        Track a moving average of this many measurements, to avoid spikes.
    """

    def __init__(self, raw=None, debug=False):
        #: last measurements for resources
        self.history = {}

        #: number of values if we take a moving average
        self.num_values = 10

        if debug:
            self.lookup_intervals = {
                'children':10,
                'cpu':1,
                'gpu':30,
                'memory':1,
                'disk':30,
                'time':1,
            }
        else:
            #: time intervals when to check resources, vs using cached values
            self.lookup_intervals = {
                'children':60,
                'cpu':10,
                'gpu':300,
                'memory':10,
                'disk':180,
                'time':10,
            }

        #: start time for resource tracking
        self.start_time = time.time()

        #: total resources controlled by the pilot
        self.total = {
            'cpu':get_cpus(),
            'gpu':get_gpus(),
            'memory':get_memory()-0.1,  # trim auto-totals to prevent going over
            'disk':get_disk()-0.1,
            'time':get_time()-0.1,
        }
        if raw:
            for r in raw:
                if (r not in self.defaults
                    or (isinstance(self.defaults[r], (float,int))
                        and not isinstance(raw[r], (float,int)))
                    or (isinstance(self.defaults[r], list)
                        and not isinstance(raw[r], list))
                    ):  # noqa: E125
                    logging.error('bad type of supplied resource: %s=%r', r, raw[r])
                else:
                    v = raw[r]
                    if isinstance(self.defaults[r], int):
                        v = int(v)
                    elif isinstance(self.defaults[r], float):
                        v = float(v)
                    elif isinstance(self.defaults[r], list):
                        v = deepcopy(v)
                    if r == 'time':
                        v -= 0.1
                    logging.info('setting %s to %r', r, v)
                    self.total[r] = v
        logging.warning('total resources: %r', self.total)

        #: available resources that are unclaimed
        self.available = deepcopy(self.total)

        #: resources allocated for each task
        self.claimed = {}  # dict of task_id:{resource}

        #: maximum usage for each claim
        # dict of task_id:{resource:{max,cnt,avg}}
        self.used = defaultdict(lambda:defaultdict(lambda:{'max':0.,'cnt':0,'avg':0.}))

        #: site info - shouldn't change while running
        self.site = get_site()

    def get_available(self):
        """
        Get available resources for new tasks to match against.

        Returns:
            dict: resources
        """
        # update available time
        self.available['time'] = self.total['time'] - (time.time()-self.start_time)/3600

        ret = deepcopy(self.available)
        for r in ret:
            if isinstance(ret[r], list):
                ret[r] = len(ret[r])
        return ret

    def get_claimed(self):
        """
        Get claimed resources for reporting purposes.

        Returns:
            dict: resources
        """
        ret = {
            'cpu':0,
            'gpu':0,
            'memory':0.,
            'disk':0.,
            'time':0,
        }
        for claim in self.claimed:
            claim_resources = self.claimed[claim]['resources']
            for r in claim_resources:
                if r == 'time':
                    ret[r] = max(ret[r],claim_resources[r])
                elif isinstance(claim_resources[r], list):
                    ret[r] += len(claim_resources[r])
                else:
                    ret[r] += claim_resources[r]
        return ret

    def claim(self, task_id, resources=None):
        """
        Attempt to claim resources for a task.

        Args:
            task_id (str): task_id that wants to use resources
            resources (dict): resources to claim (None = all)

        Raises:
            If claim cannot be satisfied, raise Exception

        Returns:
            dict: claimed resources
        """
        # update available time
        self.available['time'] = self.total['time'] - (time.time()-self.start_time)/3600
        claim = {
            'cpu':0,
            'gpu':[],
            'memory':0.,
            'disk':0.,
            'time':self.available['time'],
        }
        if not resources:
            logging.info('claiming all avaialble resources for %s', task_id)
            claim = deepcopy(self.available)
        else:
            logging.info('asking for %r', resources)
            for r in resources:
                if r not in self.available:
                    raise Exception('bad resource type: %r'%r)
                val = resources[r]
                if isinstance(self.defaults[r], (int,list)):
                    val = int(val)
                elif isinstance(self.defaults[r], float):
                    val = float(val)

                if isinstance(self.available[r], (int,float)):
                    if val > self.available[r]:
                        raise Exception('not enough {} resources available: {} > {}'.format(
                                        r, val, self.available[r]))
                    elif val > 0:
                        claim[r] = val
                elif isinstance(self.defaults[r], list):
                    if val > len(self.available[r]):
                        raise Exception('not enough {} resources available: {} > {}'.format(
                                        r, val, len(self.available[r])))
                    elif val > 0:
                        claim[r] = self.available[r][:val]

        # now that the claim is valid, remove resources from available
        for r in claim:
            if r in ('time',):  # unclaimable
                continue
            if isinstance(claim[r], (int,float)):
                self.available[r] -= claim[r]
            elif isinstance(claim[r], list):
                for v in claim[r]:
                    self.available[r].remove(v)

        self.claimed[task_id] = {
            'resources': deepcopy(claim),
            'process': None,
            'tmpdir': None,
        }

        # now send back to user
        logging.info('granted %r', claim)
        return claim

    def release(self, task_id):
        """
        Release a claim.

        Args:
            task_id (str): the task_id
        """
        if task_id not in self.claimed:
            logging.warning('release: task_id %s not in claimed', task_id)
            return
        claim = self.claimed[task_id]
        for r in claim['resources']:
            v = claim['resources'][r]
            if r in ('time',):  # unclaimable
                continue
            if isinstance(v, (int,float)):
                self.available[r] += v
            elif isinstance(v, list):
                for vv in v:
                    self.available[r].append(vv)
        del self.claimed[task_id]
        if task_id in self.used:
            del self.used[task_id]

    def register_process(self, task_id, process, tmpdir):
        """
        Register a process and tmpdir for an already claimed task.

        Args:
            task_id (str): the task_id
            process (:py:class:`psutil.Process`): a psutil process object
            tmpdir (str): temporary directory in use by process
        """
        if task_id not in self.claimed:
            logging.warning('register: task_id %s not in claimed', task_id)
            return
        self.claimed[task_id].update({
            'process': process,
            'tmpdir': tmpdir,
        })

    def check_claims(self, force=False):
        """
        Check all the claims for resource overusage.

        Args:
            force (bool): check all resources now, without caching

        Returns:
            dict: dict of task_ids:reasons that go over usage
        """
        # update available time
        self.available['time'] = self.total['time'] - (time.time()-self.start_time)/3600
        ret = {}
        for task_id in self.claimed:
            claim = self.claimed[task_id]
            try:
                usage = self.get_usage(task_id, force=force)
                logging.debug('%s is using %r', task_id, usage)
            except psutil.NoSuchProcess:
                logging.info('process has exited for %r', task_id)
                continue
            except Exception:
                logging.info('error getting usage for %r', task_id,
                             exc_info=True)
                continue
            for r in usage:
                if r == 'gpu':
                    claim_r = 100*len(claim['resources'][r])
                    avail_r = len(self.available[r])
                else:
                    claim_r = claim['resources'][r]
                    avail_r = self.available[r]
                overusage = usage[r]-claim_r
                if overusage > 0:
                    overusage_percent = usage[r]*1.0/claim_r
                    limit = self.overusage_limits[r]
                    if overusage_percent < limit['ignore']:
                        logging.info('ignoring overusage of %s for %s', r, task_id)
                    elif (r == 'time' and avail_r > 0):
                        logging.info('managable overusage of time for %s', task_id)
                    elif (r != 'time' and overusage < avail_r
                          and overusage_percent < limit['allowed']):
                        logging.info('managable overusage of %s for %s', r, task_id)
                    else:
                        ret[task_id] = 'Resource overusage for {}: {}'.format(r, usage[r])
                        break
            for r in usage:
                v = usage[r]
                u = self.used[task_id][r]
                u['avg'] = (v + u['cnt'] * u['avg'])/(u['cnt']+1)
                u['cnt'] += 1
                if v > u['max']:
                    u['max'] = v
        return ret

    def get_usage(self, task_id, force=False):
        """
        Get the resources used by this task.  Measure the process and
        its children.

        Will use cached values as necessary for speed, unless forced.

        Args:
            task_id (str): the task_id
            force (bool): check all resources now

        Returns:
            dict: measurements
        """
        if not psutil:
            raise Exception('psutil not available. cannot track resources')

        if task_id not in self.claimed:
            raise Exception('unknown claim for %s'%task_id)

        process = self.claimed[task_id]['process']
        if not process:
            raise Exception('no process to examine')

        tmpdir = self.claimed[task_id]['tmpdir']
        if not tmpdir:
            raise Exception('no tmpdir to examine')

        now = time.time()
        if task_id not in self.history:
            logging.debug('create task')
            self.history[task_id] = {
                'children': [],
                'children_last_lookup': now-100000,
                'cpu': deque(maxlen=self.num_values),
                'cpu_last_lookup': now-100000,
                'memory': deque(maxlen=self.num_values),
                'memory_last_lookup': now-100000,
                'disk': 0,
                'disk_last_lookup': now-100000,
                'gpu': 0,
                'gpu_last_lookup': now-100000,
                'create_time':process.create_time(),
                'time': 0,
            }
        task = self.history[task_id]
        logging.debug('task: %r', task)

        # recheck process children
        if force or now - task['children_last_lookup'] > self.lookup_intervals['children']:
            task['children_last_lookup'] = now
            task['children'][:] = process.children(recursive=True)
            logging.debug('children_lookup')

        lookups = {}
        for r in ('cpu','memory','disk','gpu'):
            if force or now - task[r+'_last_lookup'] > self.lookup_intervals[r]:
                task[r+'_last_lookup'] = now
                lookups[r] = True
            else:
                lookups[r] = False
        logging.debug('lookups: %r', lookups)

        # get current values
        processes = [process]+task['children']
        mem = 0
        cpu = 0
        if lookups['cpu'] or lookups['memory']:
            for p in processes:
                try:
                    with p.oneshot():
                        if lookups['cpu']:
                            cpu += p.cpu_percent()
                        if lookups['memory']:
                            mem += p.memory_info().rss
                except Exception:
                    pass
        gpu = 0
        if lookups['gpu']:
            for gpu_id in set(self.claimed[task_id]['resources']['gpu']):
                gpu_id = ''.join(filter(lambda x: x.isdigit() or x == ',', gpu_id))
                val = get_gpu_utilization_by_id(gpu_id)['utilization']
                if val != -1:
                    gpu += val
        used_resources = {
            'cpu': cpu/100.0 if cpu else None,
            'memory': mem/1000000000.0 if mem else None,
            'disk': du(tmpdir)/1000000000.0 if lookups['disk'] else None,
            'gpu': gpu/100.0 if lookups['gpu'] else None,
            'time': (now-task['create_time'])/3600,
        }
        logging.debug('used_resources: %r', used_resources)

        # now average for those that need it
        ret = {}
        for r in ('cpu','memory','disk','gpu','time'):
            if isinstance(task[r], deque):
                if used_resources[r] is not None:
                    task[r].append(used_resources[r])
                ret[r] = 0 if not task[r] else sum(task[r])/len(task[r])
            else:
                if used_resources[r] is not None:
                    task[r] = used_resources[r]
                ret[r] = task[r]
        logging.debug('ret: %r', ret)
        return ret

    def get_peak(self, task_id):
        """
        Get peak resource usage.

        Args:
            task_id (str): the task_id

        Returns:
            dict: resources
        """
        if task_id not in self.used:
            return {}
        return {r:self.used[task_id][r]['max'] for r in self.used[task_id]}

    def get_final(self, task_id):
        """
        Get final resource usage.

        Args:
            task_id (str): the task_id

        Returns:
            dict: resources
        """
        if task_id not in self.used:
            return None
        ret = {}
        for r in self.used[task_id]:
            if r == 'gpu':
                ret[r] = self.used[task_id][r]['avg']
            else:
                ret[r] = self.used[task_id][r]['max']
        return ret

    @classmethod
    def set_env(cls, resources, env=os.environ):
        """Set the environment for any resources that need it.

        Args:
            resources (dict): Resources to set
            env (dict): Environ to update (default: os.environ)
        """
        def replace(s):
            if s.startswith('CUDA'):
                s = s.replace('CUDA','')
            elif s.startswith('OCL'):
                s = s.replace('OCL','')
            return s
        if 'gpu' in resources and resources['gpu']:
            val = ','.join(replace(x) for x in set(resources['gpu']))
            env['CUDA_VISIBLE_DEVICES'] = val
            env['GPU_DEVICE_ORDINAL'] = val
            env['ROCR_VISIBLE_DEVICES'] = val
        else:
            env['CUDA_VISIBLE_DEVICES'] = '9999'
            env['GPU_DEVICE_ORDINAL'] = '9999'
            env['ROCR_VISIBLE_DEVICES'] = '9999'


def get_cpus():
    """Detect the number of available (allocated) cpus."""
    ret = None
    if os.path.exists('.machine.ad'):
        try:
            ads = classad_to_dict(open('.machine.ad').read())
        except Exception:
            pass
        else:
            if 'totalcpus' in ads:
                try:
                    ret = int(float(ads['totalcpus']))
                    logging.info('got cpus from machine ad: %r',ret)
                except Exception:
                    pass
    if (not ret) and 'NUM_CPUS' in os.environ:
        try:
            ret = int(float(os.environ['NUM_CPUS']))
            logging.info('got cpus from NUM_CPUS: %r',ret)
        except Exception:
            pass
    if not ret:
        return Resources.defaults['cpu']
    else:
        return ret


def get_gpus():
    """Detect the available (allocated) gpus.

    Returns:
        list: a list of gpus
    """
    ret = None
    if os.path.exists('.machine.ad'):
        try:
            ads = classad_to_dict(open('.machine.ad').read())
        except Exception:
            pass
        else:
            if 'assignedgpus' in ads:
                try:
                    ret = ads['assignedgpus'].strip(' "').split(',')
                    logging.info('got gpus from machine ad: %r',ret)
                except Exception:
                    pass
            for k in ads:
                if 'cuda' in k.lower() or 'ocl' in k.lower():
                    logging.warning('.machine.ad: %s=%s', k, ads[k])
    if (not ret) and 'CUDA_VISIBLE_DEVICES' in os.environ:
        try:
            ret = [x.strip() for x in os.environ['CUDA_VISIBLE_DEVICES'].split(',') if x.strip()]
            logging.info('got gpus from CUDA_VISIBLE_DEVICES: %r',ret)
        except Exception:
            pass
    if (not ret) and 'GPU_DEVICE_ORDINAL' in os.environ:
        try:
            ret = [x.strip() for x in os.environ['GPU_DEVICE_ORDINAL'].split(',') if x.strip()]
            logging.info('got gpus from GPU_DEVICE_ORDINAL: %r',ret)
        except Exception:
            pass
    if (not ret) and 'ROCR_VISIBLE_DEVICES' in os.environ:
        try:
            ret = [x.strip() for x in os.environ['ROCR_VISIBLE_DEVICES'].split(',') if x.strip()]
            logging.info('got gpus from ROCR_VISIBLE_DEVICES: %r',ret)
        except Exception:
            pass
    if (not ret) and '_CONDOR_AssignedGPUs' in os.environ:
        try:
            ret = [x.strip() for x in os.environ['_CONDOR_AssignedGPUs'].split(',') if x.strip()]
            logging.info('got gpus from _CONDOR_AssignedGPUs: %r',ret)
        except Exception:
            pass
    if (not ret) and 'NUM_GPUS' in os.environ:
        try:
            ret = [str(x) for x in range(int(os.environ['NUM_GPUS']))]
            logging.info('got gpus from NUM_GPUS: %r',ret)
        except Exception:
            pass
    if not ret:
        return deepcopy(Resources.defaults['gpu'])
    else:
        return ret


def get_memory():
    """Detect the amount of available (allocated) memory (in GB)."""
    ret = None
    if os.path.exists('.machine.ad'):
        try:
            ads = classad_to_dict(open('.machine.ad').read())
        except Exception:
            pass
        else:
            if 'totalmemory' in ads:
                try:
                    ret = float(ads['totalmemory'])/1000.
                    logging.info('got memory from machine ad: %r',ret)
                except Exception:
                    pass
    if (not ret) and 'NUM_MEMORY' in os.environ:
        try:
            ret = float(os.environ['NUM_MEMORY'])
            logging.info('got memory from NUM_MEMORY: %r',ret)
        except Exception:
            pass
    if not ret:
        return Resources.defaults['memory']
    else:
        return ret


def get_disk():
    """Detect the amount of available (allocated) disk (in GB)."""
    ret = None
    if os.path.exists('.machine.ad'):
        try:
            ads = classad_to_dict(open('.machine.ad').read())
        except Exception:
            pass
        else:
            if 'totaldisk' in ads:
                try:
                    ret = float(ads['totaldisk'])/1000000.
                    logging.info('got disk from machine ad: %r',ret)
                except Exception:
                    pass
    if (not ret) and 'NUM_DISK' in os.environ:
        try:
            ret = float(os.environ['NUM_DISK'])
            logging.info('got disk from NUM_DISK: %r',ret)
        except Exception:
            pass
    if not ret:
        return Resources.defaults['disk']
    else:
        return ret


def get_time():
    """Detect the time allocated for the job."""
    ret = None
    if os.path.exists('.machine.ad'):
        try:
            ads = classad_to_dict(open('.machine.ad').read())
        except Exception:
            logging.debug('failed to get classads from .machine.ad', exc_info=True)
        else:
            if 'pyglidein_time_to_live' in ads:
                try:
                    ret = float(ads['pyglidein_time_to_live'])/3600
                    logging.info('got pyglidein_time_to_live from machine ad: %r',ret)
                except Exception:
                    logging.info('failed to get pyglidein_time_to_live', exc_info=True)
            if (not ret) and 'glidein_max_walltime' in ads:
                try:
                    ret = float(ads['glidein_max_walltime'])/3600
                    logging.info('got glidein_max_walltime from machine ad: %r',ret)
                except Exception:
                    logging.info('failed to get glidein_max_walltime', exc_info=True)
            if (not ret) and 'timetolive' in ads:
                try:
                    ret = float(ads['timetolive'])/3600
                    logging.info('got timetolive from machine ad: %r',ret)
                except Exception:
                    logging.info('failed to get timetolive', exc_info=True)
    if (not ret) and 'NUM_TIME' in os.environ:
        try:
            ret = float(os.environ['NUM_TIME'])
            logging.info('got time from NUM_TIME: %r',ret)
        except Exception:
            pass
    if not ret:
        return Resources.defaults['time']
    else:
        return ret


def get_gpu_utilization_by_id(gpu_id):
    """Get gpu utilization based on gpu id"""
    ret = {'utilization':-1,'power':-1}
    try:
        out = subprocess.check_output(['nvidia-smi','-q','-i',str(gpu_id),'-d','UTILIZATION,POWER']).decode('utf-8')
        for line in out.split('\n'):
            line = line.strip()
            if (not line) or ':' not in line:
                continue
            k,v = [x.strip() for x in line.rsplit(':',1)]
            if k == 'Gpu':
                ret['utilization'] = int(v.replace('%','').strip())
            elif k == 'Power Draw':
                ret['power'] = float(v.replace('W','').strip())
    except Exception:
        logging.info('nvidia-smi failed for gpu %s', gpu_id, exc_info=True)
    return ret


def get_site():
    """Detect the site we are running on."""
    ret = None
    if os.path.exists('.machine.ad'):
        try:
            ads = classad_to_dict(open('.machine.ad').read())
        except Exception:
            logging.debug('failed to get classads from .machine.ad', exc_info=True)
        else:
            if 'GLIDEIN_Site' in ads:
                try:
                    ret = ads['GLIDEIN_Site']
                    logging.info('got GLIDEIN_Site from machine ad: %r',ret)
                except Exception:
                    logging.info('failed to get GLIDEIN_Site', exc_info=True)
            if (not ret) and 'GLIDEIN_SiteResource' in ads:
                try:
                    ret = ads['GLIDEIN_SiteResource']
                    logging.info('got GLIDEIN_SiteResource from machine ad: %r',ret)
                except Exception:
                    logging.info('failed to get GLIDEIN_SiteResource', exc_info=True)
            if (not ret) and 'GLIDEIN_ResourceName' in ads:
                try:
                    ret = ads['GLIDEIN_ResourceName']
                    logging.info('got GLIDEIN_ResourceName from machine ad: %r',ret)
                except Exception:
                    logging.info('failed to get GLIDEIN_ResourceName', exc_info=True)
    if (not ret) and 'GLIDEIN_Site' in os.environ:
        try:
            ret = os.environ['GLIDEIN_Site']
            logging.info('got site from GLIDEIN_Site: %r',ret)
        except Exception:
            pass
    if (not ret) and 'Site' in os.environ:
        try:
            ret = os.environ['Site']
            logging.info('got site from Site: %r',ret)
        except Exception:
            pass
    if not ret:
        return ''
    else:
        return ret


def du(path):
    """
    Perform a "du" on a path, getting the disk usage.

    Args:
        path (str): The path to analyze

    Returns:
        int: bytes used
    """
    logging.info('du of %s', path)
    total = 0
    for root,dirs,files in os.walk(path):
        for d in list(dirs):
            p = os.path.join(root,d)
            if os.path.islink(p):
                dirs.remove(d)
        for f in files:
            p = os.path.join(root,f)
            if not os.path.islink(p):
                total += os.path.getsize(p)
    logging.info('du of %s finished: %r', path, total)
    return total


def group_hasher(resources):
    """
    Hash a set of resources into a binned group.
    """
    ret = 0
    if 'cpu' in resources:
        ret = int(resources['cpu'])
    if 'gpu' in resources:
        if isinstance(resources['gpu'],(int,float)):
            ret ^= int(resources['gpu'])*100
        else:
            ret ^= len(resources['gpu'])*100
    if 'memory' in resources:
        ret ^= int(math.log(resources['memory'])*math.e)*1000
    if 'disk' in resources:
        ret ^= int(math.log(resources['disk'])*math.e)*1000000
    if 'time' in resources:
        ret ^= int(resources['time']) * 1000000000
    if 'os' in resources:
        ret ^= hash(resources['os']) & (0b11111111 << 32)
    return ret


def sanitized_requirements(reqs, use_defaults=False):
    """
    Sanitize a dict of requirements (resources) for a task.

    Args:
        reqs (dict): dict of requirements
    Returns:
        dict: sanitized requirements
    """
    reqs = {k.lower():v for k,v in reqs.items()}
    ret = {}
    all_keys = set(reqs).union(Resources.defaults)
    for k in all_keys:
        if k in reqs:
            try:
                if k in Resources.defaults:
                    if isinstance(Resources.defaults[k], (int,list)):
                        ret[k] = int(reqs[k])
                    elif isinstance(Resources.defaults[k], float):
                        ret[k] = float(reqs[k])
                else:
                    ret[k] = reqs[k]
            except ValueError:
                pass
        if use_defaults and k not in ret and k in Resources.defaults:
            if isinstance(Resources.defaults[k], (int,float,str)):
                ret[k] = Resources.defaults[k]
    return ret


def rounded_requirements(reqs, bins=None):
    """
    Round requirements into bins for submit systems to have
    fewer "choices" to consider for job matching.

    Args:
        reqs (dict): dict of requirements
        bins (dict): dict of binnings to use (None for defaults)
    Returns:
        dict: rounded requirements
    """
    if not bins:
        bins = RESOURCE_BINS

    def round_up(num, bins):
        """Round up to the next bin value"""
        for b in bins:
            if num <= b*1.05:  # within 5%
                return b
        raise Exception('num too big for bin sizes')
    ret = {}
    for k in reqs:
        v = reqs[k]
        if k in bins:
            v = round_up(v, bins[k])
        ret[k] = v
    return ret



"""
Monitor the datasets.

Send monitoring data to graphite.
"""

import argparse
import asyncio
from collections import Counter
import logging
import os

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.server.module import FakeStatsClient, StatsClientIgnoreErrors
from iceprod.server import states

logger = logging.getLogger('dataset_monitor')


TASKS_IN_PARALLEL = 100


async def process_dataset(rest_client, statsd, dataset_id):
    """
    Process a single dataset.

    Returns:
        dict: future resources
    """
    future_resources = {'gpu': 0, 'cpu': 0}
    dataset = await rest_client.request('GET', f'/datasets/{dataset_id}')
    dataset_num = dataset['dataset']
    dataset_status = dataset['status']
    jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/job_counts/status')
    jobs_counter = Counter()
    for status in jobs:
        if dataset_status in ('suspended', 'errors') and status in states.job_prev_statuses('suspended'):
            jobs_counter['suspended'] = jobs[status]
        else:
            jobs_counter[status] = jobs[status]
    for status in states.JOB_STATUS:
        if status not in jobs_counter:
            jobs_counter[status] = 0
        statsd.gauge(f'datasets.{dataset_num}.jobs.{status}', jobs_counter[status])
    tasks = await rest_client.request('GET', f'/datasets/{dataset_id}/task_counts/name_status')
    task_stats = await rest_client.request('GET', f'/datasets/{dataset_id}/task_stats')

    for name in tasks:
        tasks_counter = Counter()
        for status in tasks[name]:
            if dataset_status in ('suspended', 'errors') and status in states.task_prev_statuses('suspended'):
                tasks_counter['suspended'] += tasks[name][status]
            else:
                tasks_counter[status] = tasks[name][status]
        for status in states.TASK_STATUS:
            if status not in tasks_counter:
                tasks_counter[status] = 0
            statsd.gauge(f'datasets.{dataset_num}.tasks.{name}.{status}', tasks_counter[status])

            # now add to future resource prediction
            if status not in ('idle', 'failed', 'suspended', 'complete'):
                if name not in task_stats:
                    continue
                res = 'gpu' if task_stats[name]['gpu'] > 0 else 'cpu'
                future_resources[res] += tasks_counter[status]*task_stats[name]['avg_hrs']

    # add jobs not materialized to future resource prediction
    if dataset_status not in ('suspended', 'errors'):
        num_jobs_remaining = dataset['jobs_submitted'] - sum(jobs.values())
        for name in task_stats:
            res = 'gpu' if task_stats[name]['gpu'] > 0 else 'cpu'
            future_resources[res] += num_jobs_remaining*task_stats[name]['avg_hrs']

    return future_resources


async def run(rest_client, statsd, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        statsd (:py:class:`statsd.StatsClient`): statsd (graphite) client
        debug (bool): debug flag to propagate exceptions
    """
    try:
        future_resources = {'gpu': 0, 'cpu': 0}

        async def process_task(t):
            ret = await t
            if ret:
                for k in future_resources:
                    if k in ret:
                        future_resources[k] += ret[k]

        pending_tasks = set()
        datasets = await rest_client.request('GET', '/dataset_summaries/status')
        for status in datasets:
            for dataset_id in datasets[status]:
                while len(pending_tasks) >= TASKS_IN_PARALLEL:
                    done, pending_tasks = await asyncio.wait(pending_tasks, return_when=asyncio.FIRST_COMPLETED)
                    for t in done:
                        await process_task(t)
                pending_tasks.add(asyncio.create_task(process_dataset(rest_client, statsd, dataset_id)))

        while pending_tasks:
            done, pending_tasks = await asyncio.wait(pending_tasks)
            for t in done:
                await process_task(t)

        for res in future_resources:
            statsd.gauge(f'future_resources.{res}', int(future_resources[res]))

    except Exception:
        logger.error('error monitoring datasets', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--statsd-address', default=os.environ.get('STATSD_ADDRESS', None), help='statsd address')
    parser.add_argument('--statsd-prefix', default=os.environ.get('STATSD_PREFIX', 'site'), help='statsd site prefix')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    if not args.statsd_address:
        statsd = FakeStatsClient()
    else:
        addr = args.statsd_address
        port = 8125
        if ':' in addr:
            addr,port = addr.split(':')
            port = int(port)
        statsd = StatsClientIgnoreErrors(addr, port=port, prefix=args.statsd_prefix+'.schedule')

    asyncio.run(run(rest_client, statsd, debug=args.debug))


if __name__ == '__main__':
    main()



import logging
import json
import uuid

import pymongo
import tornado.web

from ..base_handler import APIBase
from ..auth import authorization
from iceprod.server.util import nowstr

logger = logging.getLogger('rest.pilots')


def setup(handler_cfg):
    """
    Setup method for Pilots REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            (r'/pilots', MultiPilotsHandler, handler_cfg),
            (r'/pilots/(?P<pilot_id>\w+)', PilotsHandler, handler_cfg),
        ],
        'database': 'pilots',
        'indexes': {
            'jobs': {
                'pilot_id_index': {'keys': 'pilot_id', 'unique': True},
            }
        }
    }


class MultiPilotsHandler(APIBase):
    """
    Handle multi pilots requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self):
        """
        Get pilot entries.

        Params (optional):
            queue_host: queue_host to filter by
            queue_version: queue_version to filter by
            host: host to filter by
            version: version to filter by
            keys: | separated list of keys to return for each pilot

        Returns:
            dict: {'uuid': {pilot_data}}
        """
        filters = {}
        for k in ('queue_host','queue_version','host','version'):
            tmp = self.get_argument(k, None)
            if tmp:
                filters[k] = tmp

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
            projection['pilot_id'] = True

        ret = {}
        async for row in self.db.pilots.find(filters,projection=projection):
            ret[row['pilot_id']] = row
        self.write(ret)

    @authorization(roles=['admin', 'system'])
    async def post(self):
        """
        Create a pilot entry.

        Body should contain the pilot data.

        Returns:
            dict: {'result': <pilot_id>}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'queue_host': str,
            'queue_version': str,  # iceprod version
            'resources': dict,  # min resources requested
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        # set some fields
        pilot_id = uuid.uuid1().hex
        data['pilot_id'] = pilot_id
        data['submit_date'] = nowstr()
        data['start_date'] = ''
        data['last_update'] = data['submit_date']
        if 'tasks' not in data:
            data['tasks'] = []
        if 'host' not in data:
            data['host'] = ''
        if 'site' not in data:
            data['site'] = ''
        if 'version' not in data:
            data['version'] = ''
        if 'grid_queue_id' not in data:
            data['grid_queue_id'] = ''
        if 'resources_available' not in data:
            data['resources_available'] = {}
        if 'resources_claimed' not in data:
            data['resources_claimed'] = {}

        await self.db.pilots.insert_one(data)
        self.set_status(201)
        self.write({'result': pilot_id})
        self.finish()


class PilotsHandler(APIBase):
    """
    Handle single pilot requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, pilot_id):
        """
        Get a pilot entry.

        Args:
            pilot_id (str): the pilot id

        Returns:
            dict: pilot entry
        """
        ret = await self.db.pilots.find_one({'pilot_id':pilot_id}, projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Pilot not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin', 'system'])
    async def patch(self, pilot_id):
        """
        Update a pilot entry.

        Body should contain the pilot data to update.  Note that this will
        perform a merge (not replace).

        Args:
            pilot_id (str): the pilot id

        Returns:
            dict: updated pilot entry
        """
        data = json.loads(self.request.body)
        if not data:
            raise tornado.web.HTTPError(400, reason='Missing update data')
        data['last_update'] = nowstr()

        ret = await self.db.pilots.find_one_and_update(
            {'pilot_id':pilot_id},
            {'$set':data},
            projection={'_id':False},
            return_document=pymongo.ReturnDocument.AFTER
        )
        if not ret:
            self.send_error(404, reason="Pilot not found")
        else:
            if 'site' in ret and ret['site']:
                self.module.statsd.incr('site.{}.pilot'.format(ret['site']))
            self.write(ret)
            self.finish()

    @authorization(roles=['admin', 'system'])
    async def delete(self, pilot_id):
        """
        Delete a pilot entry.

        Args:
            pilot_id (str): the pilot id

        Returns:
            dict: empty dict
        """
        ret = await self.db.pilots.find_one_and_delete({'pilot_id':pilot_id})
        if not ret:
            self.send_error(404, reason="Pilot not found")
        else:
            if 'site' in ret and ret['site']:
                self.module.statsd.incr('site.{}.pilot_delete'.format(ret['site']))
            self.write({})



"""
Server for queue management
"""
from collections import defaultdict
from functools import partial
import importlib
import logging
from pathlib import Path
import pkgutil

import motor.motor_asyncio
from rest_tools.server import RestServer
from tornado.web import RequestHandler, HTTPError
from wipac_dev_tools import from_environment

from ..s3 import boto3, S3
from ..server.module import FakeStatsClient, StatsClientIgnoreErrors
from .base_handler import IceProdRestConfig

logger = logging.getLogger('rest-server')


class Error(RequestHandler):
    def prepare(self):
        raise HTTPError(404, 'invalid api route')


class Health(RequestHandler):
    def get(self):
        self.write({})


class Server:
    def __init__(self, s3_override=None):
        default_config = {
            'HOST': 'localhost',
            'PORT': 8080,
            'DEBUG': False,
            'OPENID_URL': '',
            'OPENID_AUDIENCE': '',
            'DB_URL': 'mongodb://localhost/iceprod',
            'STATSD_ADDRESS': '',
            'STATSD_PREFIX': 'rest_api',
            'S3_ADDRESS': '',
            'S3_ACCESS_KEY': '',
            'S3_SECRET_KEY': '',
            'MAX_BODY_SIZE': 10**9,
            'ROUTE_STATS_WINDOW_SIZE': 1000,
            'ROUTE_STATS_WINDOW_TIME': 3600,
            'ROUTE_STATS_TIMEOUT': 60,
            'CI_TESTING': '',
        }
        config = from_environment(default_config)

        rest_config = {
            'debug': config['DEBUG'],
            'route_stats': {
                'window_size': config['ROUTE_STATS_WINDOW_SIZE'],
                'window_time': config['ROUTE_STATS_WINDOW_TIME'],
                'timeout': config['ROUTE_STATS_TIMEOUT'],
            }
        }
        if config['OPENID_URL']:
            logging.info(f'enabling auth via {config["OPENID_URL"]} for aud "{config["OPENID_AUDIENCE"]}"')
            rest_config.update({
                'auth': {
                    'openid_url': config['OPENID_URL'],
                    'audience': config['OPENID_AUDIENCE'],
                }
            })
        elif config['CI_TESTING']:
            rest_config.update({
                'auth': {
                    'secret': 'secret',
                }
            })
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        statsd = FakeStatsClient()
        if config['STATSD_ADDRESS']:
            try:
                addr = config['STATSD_ADDRESS']
                port = 8125
                if ':' in addr:
                    addr,port = addr.split(':')
                    port = int(port)
                statsd = StatsClientIgnoreErrors(addr, port=port, prefix=config['STATSD_PREFIX'])
            except Exception:
                logger.warning('failed to connect to statsd: %r', config['STATSD_ADDRESS'], exc_info=True)

        s3conn = None
        if s3_override:
            logging.warning('S3 in testing mode')
            s3conn = S3('', '', '', mock_s3=s3_override)
        elif boto3 and config['S3_ACCESS_KEY'] and config['S3_SECRET_KEY']:
            s3conn = S3(config['S3_ADDRESS'], config['S3_ACCESS_KEY'], config['S3_SECRET_KEY'])
        else:
            logger.warning('S3 is not available!')

        logging_url = config["DB_URL"].split('@')[-1] if '@' in config["DB_URL"] else config["DB_URL"]
        logging.info(f'DB: {logging_url}')
        db_url, db_name = config['DB_URL'].rsplit('/', 1)
        self.db = motor.motor_asyncio.AsyncIOMotorClient(db_url)
        logging.info(f'DB name: {db_name}')
        self.indexes = defaultdict(partial(defaultdict, dict))

        kwargs = IceProdRestConfig(rest_config, statsd=statsd, database=self.db, s3conn=s3conn)

        server = RestServer(debug=config['DEBUG'], max_body_size=config['MAX_BODY_SIZE'])

        handler_path = str(Path(__file__).parent / 'handlers')
        for _, name, _ in pkgutil.iter_modules([handler_path]):
            mod = importlib.import_module(f'iceprod.rest.handlers.{name}')
            ret = mod.setup(kwargs)
            for route,cls,kw in ret['routes']:
                kw2 = kw.copy()
                kw2['database'] = kw['database'][ret['database']]
                kw2['auth_database'] = kw['database']['auth']
                server.add_route(route, cls, kw2)
            ii = self.indexes[ret['database']]
            for col in ret['indexes']:
                ii[col].update(ret['indexes'][col])

        server.add_route('/healthz', Health)
        server.add_route(r'/(.*)', Error)

        server.startup(address=config['HOST'], port=config['PORT'])

        self.server = server

    async def start(self):
        for database in self.indexes:
            db = self.db[database]
            for collection in self.indexes[database]:
                existing = await db[collection].index_information()
                for name in self.indexes[database][collection]:
                    if name not in existing:
                        kwargs = self.indexes[database][collection][name]
                        logging.info('DB: creating index %s/%s:%s %r', database, collection, name, kwargs)
                        await db[collection].create_index(name=name, **kwargs)

    async def stop(self):
        await self.server.stop()



import asyncio
import json
import logging
import time

from cachetools.func import ttl_cache
import httpx
import jwt
from rest_tools.utils.auth import OpenIDAuth

logger = logging.getLogger('refresh_service')


@ttl_cache(maxsize=256, ttl=3600)
def get_auth(url):
    return OpenIDAuth(url)


def get_expiration(token):
    """
    Find a token's expiration time.

    Args:
        token (str): jwt token
    Returns:
        float: expiration unix time
    """
    return jwt.decode(token, options={"verify_signature": False})['exp']


def is_expired(cred):
    """
    Check if an OAuth credential is expired.

    Will mark credential as expired if the access token has less than 5 seconds left.

    Args:
        cred (dict): credential dict
    Returns:
        bool: True if expired
    """
    if cred['type'] != 'oauth':
        return False
    return cred['expiration'] < (time.time() + 5)


class RefreshService:
    """
    OAuth refresh service

    Args:
        database (:motor.motor_asyncio.AsyncIOMotorClient:): mongo database
        clients (str): json string of {url: [client id, client secret]}
        refresh_window (float): how long after last use to keep refreshing (in hours)
        expire_buffer (float): how long before expiration to refresh (in hours)
        service_run_interval (float): seconds between refresh runs
    """
    def __init__(self, database, clients, refresh_window, expire_buffer, service_run_interval):
        self.db = database
        self.clients = json.loads(clients)
        self.refresh_window = refresh_window * 3600
        self.expire_buffer = expire_buffer * 3600

        self.start_time = time.time()
        self.service_run_interval = service_run_interval
        self.last_run_time = None
        self.last_success_time = None

    async def refresh_cred(self, cred):
        if not cred.get('refresh_token'):
            raise Exception('cred does not have a refresh token')

        openid_url = jwt.decode(cred['refresh_token'], options={"verify_signature": False})['iss']
        if openid_url not in self.clients:
            raise Exception('jwt issuer not registered')
        auth = get_auth(openid_url)

        # try the refresh token
        args = {
            'grant_type': 'refresh_token',
            'refresh_token': cred['refresh_token'],
            'client_id': self.clients[openid_url][0],
        }
        if len(self.clients[openid_url]) > 1:
            args['client_secret'] = self.clients[openid_url][1]

        new_cred = {}
        try:
            async with httpx.AsyncClient() as client:
                r = await client.post(auth.token_url, data=args)
            r.raise_for_status()
            req = r.json()
        except httpx.HTTPError as exc:
            logger.debug('%r', exc.response.text)
            try:
                req = exc.response.json()
            except Exception:
                req = {}
            error = req.get('error', '')
            raise Exception(f'Refresh request failed: {error}') from exc
        else:
            logger.debug('OpenID token refreshed')
            new_cred['access_token'] = req['access_token']
            new_cred['refresh_token'] = req['refresh_token']
            new_cred['expiration'] = get_expiration(req['access_token'])
            logger.debug('%r', new_cred)

        return new_cred

    def should_refresh(self, cred):
        now = time.time()
        refresh_exp = now + self.expire_buffer
        last_use_date = now - self.refresh_window

        if cred.get('last_use', 0) < last_use_date:
            return False
        return cred['expiration'] < refresh_exp

    async def run(self):
        """
        Run loop.
        """
        while True:
            await self._run_once()
            await asyncio.sleep(self.service_run_interval)

    async def _run_once(self):
        try:
            self.last_run_time = time.time()
            last_use_check = self.last_run_time - self.refresh_window
            exp_check = self.last_run_time + self.expire_buffer
            filters = {
                'type': 'oauth',
                'last_use': {'$gt': last_use_check},
                'expiration': {'$lt': exp_check},
            }

            user_creds = {}
            async for row in self.db.user_creds.find(filters, {'_id': False}):
                user_creds[row['url']] = row

            for cred in user_creds.values():
                if cred['type'] != 'oauth':
                    continue
                if not cred['refresh_token']:
                    logger.info('skipping non-refresh token for user %s, url %s', cred['username'], cred['url'])
                    continue
                logger.debug('cred: %r', cred)
                try:
                    if self.should_refresh(cred):
                        args = await self.refresh_cred(cred)
                        await self.db.user_creds.update_one({'username': cred['username'], 'url': cred['url']}, {'$set': args})
                        logger.info('refreshed token for user %s, url %s', cred['username'], cred['url'])
                    else:
                        logger.info('not yet time to refresh token for user %s, url %s', cred['username'], cred['url'])
                except Exception:
                    logger.error('error refreshing token for user %s, url: %s', cred['username'], cred['url'], exc_info=True)

            group_creds = {}
            async for row in self.db.group_creds.find(filters, {'_id': False}):
                group_creds[row['url']] = row

            for cred in group_creds.values():
                if cred['type'] != 'oauth':
                    continue
                if not cred['refresh_token']:
                    logger.info('skipping non-refresh token for group %s, url %s', cred['groupname'], cred['url'])
                    continue
                try:
                    if self.should_refresh(cred):
                        args = await self.refresh_cred(cred)
                        await self.db.group_creds.update_one({'groupname': cred['groupname'], 'url': cred['url']}, {'$set': args})
                        logger.info('refreshed token for group %s, url %s', cred['groupname'], cred['url'])
                    else:
                        logger.info('not yet time to refresh token for group %s, url %s', cred['groupname'], cred['url'])
                except Exception:
                    logger.error('error refreshing token for group %s, url: %s', cred['groupname'], cred['url'], exc_info=True)

            self.last_success_time = time.time()
        except Exception:
            logger.error('error running refresh', exc_info=True)



"""
Task submission to condor, using condor file transfer.
"""
import os
import logging

from iceprod.server.plugins.condor_direct import condor_direct

logger = logging.getLogger('plugin-condor_file_transfer')


class condor_file_transfer(condor_direct):
    """Plugin Overrides for HTCondor with file transfer for osdf://"""

    batch_site = 'CondorFileTransfer'
    batch_resources = {}

    async def customize_task_config(self, task_cfg, **kwargs):
        """Do OSDF file transfers via condor, so move files to batchsys"""
        await super().customize_task_config(task_cfg, **kwargs)

        in_files = []
        out_files = []

        def escape_remap(x):
            return x.replace('=','\\=').replace(';','\\;')

        def process_data(cfg):
            new_data = []
            for d in cfg.get('data', []):
                if d['remote'].startswith('osdf://'):
                    if d['movement'] in ('input', 'both'):
                        in_files.append(d['remote'])
                    if d['movement'] in ('output', 'both'):
                        local = d['local'] if d['local'] else os.path.basename(d['remote'])
                        remote = d['remote']
                        out_files.append(f'{escape_remap(local)} = {escape_remap(remote)}')
                else:
                    new_data.append(d)
            cfg['data'] = new_data

        process_data(task_cfg)
        for tray in task_cfg['trays']:
            process_data(tray)
            for module in tray['modules']:
                process_data(module)

        if in_files or out_files:
            # batchsys config can be None in config!
            batchsys = task_cfg.get('batchsys', None)
            if not batchsys:
                batchsys = {}
            batchsys_condor = batchsys.get('condor', {})
            reqs = batchsys_condor.get('requirements','')
            if reqs:
                reqs += ' && regexp("osdf",HasFileTransferPluginMethods)'
            else:
                reqs = 'regexp("osdf",HasFileTransferPluginMethods)'
            batchsys_condor['requirements'] = reqs
            batchsys_condor['transfer_input_files'] = ','.join(in_files)
            batchsys_condor['transfer_output_files'] = ','.join(v.split('=',1)[0].strip() for v in out_files)
            batchsys_condor['transfer_output_remaps'] = ','.join(out_files)
            batchsys['condor'] = batchsys_condor
            task_cfg['batchsys'] = batchsys



#!/usr/bin/env python
"""
Set gridftp proxy.
"""

from __future__ import absolute_import, division, print_function

import os
import sys
import logging
import argparse
import getpass

# Assuming that the directory structure remains constant, this will automatically set the python path to find iceprod
bin_dir = os.path.dirname( os.path.abspath(sys.argv[0]) )
root_path = os.path.dirname( bin_dir )
if not root_path in sys.path: sys.path.append( root_path )

from iceprod.server.gridftp import SiteGlobusProxy

def main():
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description='IceProd Gridftp Proxy Helper')
    parser.add_argument('-f','--cfgfile',dest='file',type=str,default=None,
                        help='Path to gridftp config file')
    args = parser.parse_args()

    p = SiteGlobusProxy(cfgfile=args.file)
    passphrase = getpass.getpass('passphrase: ')
    p.set_passphrase(passphrase)
    try:
        p.update_proxy()
    except Exception as e:
        logging.error('proxy error', exc_info=True)

if __name__ == '__main__':
    main()
